import datetime
import math
import os
from dataclasses import dataclass, field
from textwrap import dedent
from typing import Optional, Tuple, Union

import cv2
import numpy as np
from moviepy.editor import VideoFileClip
from rich import print
from rich.console import Console
from rich.progress import Progress

import videoxt.constants as C
import videoxt.utils as utils
import videoxt.validators as V


@dataclass
class BaseVideoExtractor:
    video_path: str
    start_time: Union[float, int, str] = 0.0
    stop_time: Optional[Union[float, int, str]] = None
    fps: Optional[float] = None
    dimensions: Optional[Tuple[int, int]] = None
    resize: float = 1.0
    rotate: int = 0
    output_dir: Optional[str] = None
    output_filename: Optional[str] = None
    quiet: bool = False
    emoji: bool = False
    extraction_type: str = field(init=False)
    start_second: float = field(init=False)
    stop_second: float = field(init=False)
    start_frame: float = field(init=False)
    stop_frame: float = field(init=False)
    target_dimensions: Tuple[int, int] = field(init=False)
    frame_count: float = field(init=False)
    video_abspath: str = field(init=False)
    video_basename: str = field(init=False)
    video_dirname: str = field(init=False)
    video_filename: str = field(init=False)
    video_length: str = field(init=False)
    video_length_seconds: float = field(init=False)
    video_dimensions: Tuple[int, int] = field(init=False)

    def __post_init__(self) -> None:
        """The order in which these methods are called is important."""
        self.extraction_type = self.__class__.__name__.lower().replace("videoto", "")
        self._validate_user_input()
        self._initialize_video_attributes()
        self._initialize_video_metadata()
        self._set_video_extraction_seconds_range()
        V.validate_video_extraction_range(
            self.start_second, self.stop_second, self.video_length_seconds
        )
        self._set_video_extraction_frame_range()
        self._set_default_output_dir()
        self._set_default_output_filename()
        self._set_target_dimensions()

    def _validate_user_input(self) -> None:
        """Validations when run from command-line are handled by argparse."""
        if not C.IS_TERMINAL:
            self.video_path = V.valid_filepath(self.video_path)
            self.start_time = V.valid_start_time(self.start_time)
            self.stop_time = (
                V.valid_stop_time(self.stop_time)
                if self.stop_time is not None
                else self.stop_time
            )
            self.fps = V.positive_float(self.fps) if self.fps is not None else None
            self.resize = V.valid_resize_value(self.resize)
            self.dimensions = (
                V.valid_dimensions(self.dimensions)
                if self.dimensions is not None
                else self.dimensions
            )
            self.rotate = V.valid_rotate_value(self.rotate)
            self.output_dir = (
                V.valid_dir(self.output_dir)
                if self.output_dir is not None
                else self.output_dir
            )
            if self.output_filename is not None:
                self.output_filename = V.valid_filename(self.output_filename)

    def _initialize_video_attributes(self) -> None:
        self.video_abspath = (
            self.video_path
            if os.path.isabs(self.video_path)
            else os.path.abspath(self.video_path)
        )
        self.video_dirname = os.path.dirname(self.video_abspath)
        self.video_basename = os.path.basename(self.video_abspath)
        self.video_filename = os.path.splitext(self.video_basename)[0]

    def _initialize_video_metadata(self) -> None:
        """Video metadata initialized: frame count, fps, video length, and video dimensions."""
        video_capture = cv2.VideoCapture(self.video_abspath)

        self.frame_count = round(video_capture.get(cv2.CAP_PROP_FRAME_COUNT), 2)

        if self.fps is None:
            self.fps = round(video_capture.get(cv2.CAP_PROP_FPS), 2)

        self.video_length_seconds = round(self.frame_count / self.fps, 2)
        self.video_length = str(
            datetime.timedelta(seconds=self.video_length_seconds)
        ).split(".")[0]

        self.video_dimensions = (
            int(video_capture.get(cv2.CAP_PROP_FRAME_WIDTH)),
            int(video_capture.get(cv2.CAP_PROP_FRAME_HEIGHT)),
        )

        video_capture.release()

    def _set_video_extraction_seconds_range(self) -> None:
        """Set the start and stop seconds for video extraction

        The stop second is set to the video length if the stop time is not specified or
        if the stop second is greater than the video length.
        """
        self.start_second = (
            utils.timestamp_to_seconds(self.start_time)
            if isinstance(self.start_time, str)
            else self.start_time
        )

        if self.stop_time is not None:
            self.stop_second = (
                utils.timestamp_to_seconds(self.stop_time)
                if isinstance(self.stop_time, str)
                else self.stop_time
            )
            self.stop_second = (
                self.video_length_seconds
                if self.stop_second > self.video_length_seconds
                else self.stop_second
            )
        else:
            self.stop_second = self.video_length_seconds

    def _set_video_extraction_frame_range(self) -> None:
        """Set the start and stop frame for video extraction

        The stop frame is set to the frame count if the stop time is not specified or
        if the stop frame is greater than the frame count.
        """
        self.start_frame = round(self.start_second * self.fps, 2)

        if self.stop_time is not None:
            self.stop_frame = round(self.stop_second * self.fps, 2)
            self.stop_frame = (
                self.frame_count
                if self.stop_frame > self.frame_count
                else self.stop_frame
            )
        else:
            self.stop_frame = self.frame_count

    def _set_default_output_dir(self) -> None:
        """Set the default output directory for extracted frames or gif.

        For the 'images' extraction type:
        If the output directory is not specified, the default directory will be created
        in the same directory as the video file. If the default directory already exists,
        a new directory will be created with a number appended to the end of the name.

        For the 'gif' extraction type:
        If the output directory is not specified, the gif will be saved in the same directory
        as the video file.
        """
        if self.extraction_type == "images":
            if self.output_dir is None:
                output_dir_name = f"{self.video_filename}_frames"
                output_dir = os.path.join(self.video_dirname, output_dir_name)

                if os.path.exists(output_dir):
                    index = 2
                    while os.path.exists(output_dir):
                        output_dir = os.path.join(
                            self.video_dirname, f"{output_dir_name} ({index})"
                        )
                        index += 1

                self.output_dir = output_dir

        elif self.extraction_type == "gif":
            if self.output_dir is None:
                self.output_dir = self.video_dirname

    def _set_default_output_filename(self) -> None:
        """Set the default output filename for extracted frames or gif.

        For the 'images' extraction type:
        If the output filename is not specified, the default filename will be the name of the video file
        with '_frame' appended to the end.

        For the 'gif' extraction type:
        If the output filename is not specified, the default filename will be the name of the video file
        with '.gif' appended to the end. If the output filename is specified, the '.gif' extension will be
        appended to the end of the filename if it is not already present.
        """
        if self.extraction_type == "images":
            if self.output_filename is None:
                self.output_filename = f"{self.video_filename}_frame"

        elif self.extraction_type == "gif":
            if self.output_filename is None:
                self.output_filename = f"{self.video_filename}.gif"
            else:
                if not self.output_filename.endswith(".gif"):
                    self.output_filename += ".gif"

    def _set_target_dimensions(self) -> None:
        """Set the target dimensions for extracted frames or gif.

        If the resize value is not 1, each dimension will be multiplied by the resize value,
        regardless of whether or not the user specified the dimensions.
        If the user does not specify the dimensions, the target dimensions
        will be the same as the video dimensions.
        """
        if self.dimensions is not None:
            if self.resize != 1:
                self.target_dimensions = tuple(
                    [int(dimension * self.resize) for dimension in self.dimensions]
                )
            else:
                self.target_dimensions = self.dimensions
        else:
            if self.resize != 1:
                self.target_dimensions = tuple(
                    [
                        int(dimension * self.resize)
                        for dimension in self.video_dimensions
                    ]
                )
            else:
                self.target_dimensions = self.video_dimensions

    def __str__(self) -> str:
        start_time_display = f"{utils.seconds_to_timestamp(int(self.start_second))} | {self.start_second} seconds"
        stop_time_display = f"{utils.seconds_to_timestamp(int(self.stop_second))} | {self.stop_second} seconds"
        if self.emoji:
            start_time_display = f"{C.EMOJI_MAP['start']} {start_time_display}"
            stop_time_display = f"{C.EMOJI_MAP['stop']} {stop_time_display}"

        return dedent(
            f"""
            {'VIDEO' if not self.emoji else C.EMOJI_MAP['video']}
              input:        {self.video_path!r}
              length:       {self.video_length} | {self.video_length_seconds} seconds
              fps:          {self.fps}
              frame count:  {self.frame_count}
              dimensions:   {self.video_dimensions}
            {'EXTRACTION' if not self.emoji else C.EMOJI_MAP['extraction']}
              type:         {self.extraction_type!r}
              start time:   {start_time_display}
              stop time:    {stop_time_display}
              start frame:  {self.start_frame + 1 if not self.emoji else C.EMOJI_MAP['start'] + ' ' + str(self.start_frame + 1)}
              stop frame:   {self.stop_frame + 1 if not self.emoji else C.EMOJI_MAP['stop'] + ' ' + str(self.stop_frame + 1)}
              output dir:   {self.output_dir}"""
        )


@dataclass
class VideoToImages(BaseVideoExtractor):
    capture_rate: int = 1
    image_format: str = "jpg"
    images_expected: int = field(init=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        self._validate_image_input()
        self._set_images_expected()

    def _validate_image_input(self) -> None:
        if not C.IS_TERMINAL:
            self.image_format = V.valid_image_format(self.image_format)
            self.capture_rate = V.positive_int(self.capture_rate)

        self.capture_rate = V.valid_capture_rate(
            self.capture_rate, self.start_frame, self.stop_frame
        )

    def _set_images_expected(self) -> None:
        self.images_expected = math.ceil(
            (self.stop_frame - self.start_frame) / self.capture_rate
        )

    def _apply_image_transformations(self, image: np.ndarray) -> np.ndarray:
        image = cv2.resize(image, self.target_dimensions)

        if self.rotate in C.ROTATION_MAP:
            image = cv2.rotate(image, C.ROTATION_MAP[self.rotate])

        return image

    def extract_images(self) -> None:
        video_capture = cv2.VideoCapture(self.video_path)
        frame_position = int(self.start_frame)
        video_capture.set(cv2.CAP_PROP_POS_FRAMES, frame_position)

        # Only creates the directory if the default output directory `video_frames_*.jpg` was used.
        if not os.path.exists(str(self.output_dir)):
            os.mkdir(str(self.output_dir))

        if not self.quiet:
            print(str(self))

        with Progress(transient=True) as progress:
            extract_task = progress.add_task(
                f"[yellow]EXTRACTING FRAMES {self.video_basename}",
                total=self.images_expected,
            )

            images_written = 1
            while video_capture.isOpened():
                read_successful, image = video_capture.read()
                if read_successful:
                    # Write image to disk,
                    image_filename = f"{self.output_filename}_{frame_position + 1}.{self.image_format}"
                    image_path = os.path.join(str(self.output_dir), image_filename)
                    image = self._apply_image_transformations(image)
                    cv2.imwrite(image_path, image)

                    if self.images_expected == images_written:
                        video_capture.release()
                        break

                    # cv2.VideoCapture.read() increments and sets the next frame position by 1 on next read.
                    # Therefore, the frame position is incremented by the capture rate only if capture rate > 1.
                    frame_position += self.capture_rate
                    if self.capture_rate != 1:
                        video_capture.set(cv2.CAP_PROP_POS_FRAMES, frame_position)

                    progress.update(
                        extract_task,
                        advance=1,
                        description=f"[yellow]EXTRACTING FRAMES FROM {self.video_basename!r} [{images_written}/{self.images_expected}]",
                    )

                    images_written += 1

                else:
                    video_capture.release()
                    break

        print(
            f"[green]EXTRACTION COMPLETE [{images_written}] IMAGES SAVED: {self.output_dir}[/green]"
        )

    def __str__(self) -> str:
        if self.emoji:
            if self.resize < 1:
                resize_display = str(self.resize) + " " + C.EMOJI_MAP["resize_small"]
            elif self.resize > 1:
                resize_display = str(self.resize) + " " + C.EMOJI_MAP["resize_large"]
            else:
                resize_display = str(self.resize) + " " + C.EMOJI_MAP["resize_normal"]

            rotate_display = str(self.rotate) + " " + C.EMOJI_MAP[self.rotate]
        else:
            resize_display = str(self.resize)
            rotate_display = str(self.rotate)

        name_display = f"{self.output_filename}_*.{self.image_format}"

        return super().__str__() + dedent(
            f"""
            {"IMAGES" if not self.emoji else C.EMOJI_MAP['image']}
              filenames:    {name_display!r}
              format:       {self.image_format!r}
              resize:       {resize_display}
              rotate:       {rotate_display}
              capture rate: {self.capture_rate}
              expected:     {self.images_expected}
              dimensions:   {self.target_dimensions}
              """
        )


@dataclass
class VideoToGIF(BaseVideoExtractor):
    speed: float = 1.0

    def __post_init__(self) -> None:
        super().__post_init__()
        self._validate_gif_input()

    def _validate_gif_input(self) -> None:
        if not C.IS_TERMINAL:
            self.speed = V.positive_float(self.speed)

    def _apply_gif_transformations(self, subclip: VideoFileClip) -> VideoFileClip:
        subclip = subclip.resize(self.target_dimensions).speedx(self.speed)

        if self.rotate != 0:
            subclip = subclip.rotate(self.rotate)

        return subclip

    def create_gif(self) -> None:
        with VideoFileClip(self.video_path, audio=False) as clip:
            # Create a subclip of the video from start_second to stop_second and apply transformations.
            subclip = clip.subclip(self.start_second, self.stop_second)
            subclip = self._apply_gif_transformations(subclip)

            gif_path = os.path.join(str(self.output_dir), str(self.output_filename))

            if not self.quiet:
                print(str(self))

            with Console().status(
                f"[yellow]CREATING GIF {self.output_filename!r} FROM {self.video_basename!r}[/yellow]"
            ):
                subclip.write_gif(
                    gif_path,
                    fps=self.fps,
                    logger=None,
                )
            print(f"[green]GIF CREATED HERE: {gif_path}[/green]")

    def __str__(self) -> str:
        if self.emoji:
            if self.resize < 1:
                resize_display = str(self.resize) + " " + C.EMOJI_MAP["resize_small"]
            elif self.resize > 1:
                resize_display = str(self.resize) + " " + C.EMOJI_MAP["resize_large"]
            else:
                resize_display = str(self.resize) + " " + C.EMOJI_MAP["resize_normal"]

            rotate_display = str(self.rotate) + " " + C.EMOJI_MAP[self.rotate]

            if self.speed > 1:
                speed_display = str(self.speed) + " " + C.EMOJI_MAP["speed_fast"]
            elif self.speed < 1:
                speed_display = str(self.speed) + " " + C.EMOJI_MAP["speed_slow"]
            else:
                speed_display = str(self.speed) + " " + C.EMOJI_MAP["speed_normal"]
        else:
            resize_display = str(self.resize)
            rotate_display = str(self.rotate)
            speed_display = str(self.speed)

        return super().__str__() + dedent(
            f"""
            {"GIF" if not self.emoji else C.EMOJI_MAP['gif']}
              filename:     {self.output_filename!r}
              resize:       {resize_display}
              rotate:       {rotate_display}
              speed:        {speed_display}
              dimensions:   {self.target_dimensions}
              """
        )



import importlib.metadata

import cv2

VERSION = importlib.metadata.version("videoxt")

IS_TERMINAL = False

VALID_ROTATE_VALUES = (0, 90, 180, 270)

ROTATION_MAP = {
    90: cv2.ROTATE_90_CLOCKWISE,
    180: cv2.ROTATE_180,
    270: cv2.ROTATE_90_COUNTERCLOCKWISE,
}

VALID_IMAGE_FORMATS = (
    "bmp",
    "dib",
    "jpeg",
    "jpg",
    "png",
    "tiff",
    "tif",
    "webp",
)

EMOJI_MAP = {
    "video": "VIDEO \N{FILM PROJECTOR}",
    "extraction": "EXTRACTION \N{TOOTH}",
    "image": "IMAGES \N{CAMERA WITH FLASH}",
    "gif": "GIF \N{FILM FRAMES}",
    "resize_small": "\N{BUG}",
    "resize_normal": "\N{GOAT}",
    "resize_large": "\N{WHALE}",
    "speed_slow": "\N{TURTLE}",
    "speed_normal": "\N{MONKEY}",
    "speed_fast": "\N{RABBIT}",
    "stop": "\N{RAISED HAND}",
    "start": "\N{CLAPPER BOARD}",
    0: "\N{UPWARDS BLACK ARROW}",
    90: "\N{BLACK RIGHTWARDS ARROW}",
    180: "\N{DOWNWARDS BLACK ARROW}",
    270: "\N{LEFTWARDS BLACK ARROW}",
}



import argparse
import sys
from typing import Optional, Sequence

from rich import print

import videoxt.constants as C
import videoxt.validators as V
from videoxt.extractors import VideoToGIF, VideoToImages


def main(argv: Optional[Sequence[str]] = None) -> None:
    C.IS_TERMINAL = True

    # parent_parser houses shared subparser args
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument(
        "video_path",
        type=V.valid_filepath,
        help="Path to the video file with extension.",
    )
    parent_parser.add_argument(
        "--start-time",
        type=V.valid_start_time,
        default=0.0,
        metavar="",
        dest="start_time",
        help="Specify the start time of the video to extract frames from. Default is the start of the video.",
    )
    parent_parser.add_argument(
        "--stop-time",
        type=V.valid_stop_time,
        metavar="",
        dest="stop_time",
        help="Specify the stop time of the video to extract frames from. Default is the end of the video.",
    )
    parent_parser.add_argument(
        "--fps",
        type=V.positive_float,
        metavar="",
        help="Specify the video frames per second. This will override the video metadata FPS.",
    )
    parent_parser.add_argument(
        "--dimensions",
        type=V.positive_int,
        nargs=2,
        metavar="",
        help="Specify the media output dimensions as space-separated values (Ex: --dimensions 1920 1080). Defaults to native video dimensions.",
    )
    parent_parser.add_argument(
        "--resize",
        type=V.valid_resize_value,
        default=1.0,
        metavar="",
        help="Resize the media output by a factor of n. Default is 1.0, which is the original size of the video.",
    )
    parent_parser.add_argument(
        "--rotate",
        type=V.valid_rotate_value,
        default=0,
        metavar="",
        dest="rotate",
        help="Rotate the media output by 90, 180, or 270 degrees. Default is 0 (no rotation).",
    )
    parent_parser.add_argument(
        "--output-dir",
        type=V.valid_dir,
        metavar="",
        dest="output_dir",
        help="Directory to save the media output to. Default is within input video directory.",
    )
    parent_parser.add_argument(
        "--output-filename",
        type=V.valid_filename,
        metavar="",
        dest="output_filename",
        help="Specify the file name of the media output. Default is the input video file name.",
    )
    parent_parser.add_argument(
        "--quiet",
        action="store_true",
        help="Disable extraction details in terminal.",
    )
    parent_parser.add_argument(
        "--emoji",
        action="store_true",
        help="Enable emoji's in terminal.",
    )

    # main parser and subparsers
    main_parser = argparse.ArgumentParser(
        prog="vxt",
        description="Extract individual frames from a video, or create a GIF between two points in a video.",
    )
    main_parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {C.VERSION}",
    )
    subparsers = main_parser.add_subparsers(dest="subparser_name", required=True)

    # img subparser
    parser_img = subparsers.add_parser(
        "images",
        help="Extract individual frames from a video as images.",
        parents=[parent_parser],
    )
    parser_img.add_argument(
        "--capture-rate",
        type=V.positive_int,
        default=1,
        metavar="",
        dest="capture_rate",
        help="Capture every nth video frame. Default is 1, which captures every frame.",
    )
    parser_img.add_argument(
        "--image-format",
        type=V.valid_image_format,
        default="jpg",
        metavar="",
        dest="image_format",
        help="Specify the image format to save the frames as. Default is jpg.",
    )

    # gif subparser
    parser_gif = subparsers.add_parser(
        "gif",
        help="create a GIF between two points in a video.",
        parents=[parent_parser],
    )
    parser_gif.add_argument(
        "--speed",
        type=V.positive_float,
        default=1.0,
        metavar="",
        help="Speed of the GIF animation. Default is 1.0, which is the original speed of the video.",
    )

    # parse args
    args = main_parser.parse_args(argv)

    if args.subparser_name == "images":
        try:
            vti = VideoToImages(
                video_path=args.video_path,
                start_time=args.start_time,
                stop_time=args.stop_time,
                dimensions=args.dimensions,
                resize=args.resize,
                rotate=args.rotate,
                fps=args.fps,
                output_dir=args.output_dir,
                output_filename=args.output_filename,
                quiet=args.quiet,
                emoji=args.emoji,
                capture_rate=args.capture_rate,
                image_format=args.image_format,
            )
        except argparse.ArgumentTypeError as e:
            print(e)
            sys.exit(1)
        else:
            vti.extract_images()

    if args.subparser_name == "gif":
        try:
            vtg = VideoToGIF(
                video_path=args.video_path,
                start_time=args.start_time,
                stop_time=args.stop_time,
                dimensions=args.dimensions,
                resize=args.resize,
                rotate=args.rotate,
                fps=args.fps,
                output_dir=args.output_dir,
                output_filename=args.output_filename,
                quiet=args.quiet,
                emoji=args.emoji,
                speed=args.speed,
            )
        except argparse.ArgumentTypeError as e:
            print(e)
            sys.exit(1)
        else:
            vtg.create_gif()


if __name__ == "__main__":
    main()



import argparse
import os
import re
from typing import Tuple, Union

import videoxt.constants as C
import videoxt.utils as utils


class ValidationException(Exception):
    pass


def positive_int(num: Union[int, str]) -> int:
    error_msg_value = f"expected integer, got {num!r}"

    try:
        value = int(num)
    except ValueError:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg_value)
        raise ValidationException(error_msg_value)

    error_msg_positive = f"expected positive integer, got {num}"
    if value <= 0:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg_positive)
        raise ValidationException(error_msg_positive)

    return value


def positive_float(num: Union[float, str]) -> float:
    error_msg_value = f"expected numeric value, got {num!r}"

    try:
        value = float(num)
    except ValueError:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg_value)
        raise ValidationException(error_msg_value)

    error_msg_positive = f"expected positive number, got {num}"

    if value <= 0:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg_positive)
        raise ValidationException(error_msg_positive)

    return value


def non_negative_int(num: Union[int, str]) -> int:
    error_msg_value = f"expected integer, got {num!r}"

    try:
        value = int(num)
    except ValueError:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg_value)
        raise ValidationException(error_msg_value)

    error_msg_non_negative = f"expected non-negative integer, got {num}"

    if value < 0:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg_non_negative)
        raise ValidationException(error_msg_non_negative)

    return value


def non_negative_float(num: Union[int, float, str]) -> float:
    error_msg_value = f"expected numeric value, got {num!r}"

    try:
        value = float(num)
    except ValueError:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg_value)
        raise ValidationException(error_msg_value)

    error_msg_non_negative = f"expected non-negative number, got {num}"

    if value < 0:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg_non_negative)
        raise ValidationException(error_msg_non_negative)

    return value


def valid_filepath(filepath: str) -> str:
    error_msg = f"file not found, got {filepath!r}"

    if not os.path.isfile(filepath):
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg)
        raise ValidationException(error_msg)

    return filepath


def valid_dir(dir: str) -> str:
    error_msg_not_found = f"directory not found, got {dir!r}"

    if not os.path.exists(dir):
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg_not_found)
        raise ValidationException(error_msg_not_found)

    error_msg_not_dir = f"expected directory, got {dir!r}"

    if not os.path.isdir(dir):
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg_not_dir)
        raise ValidationException(error_msg_not_dir)

    return dir


def valid_filename(filename: str) -> str:
    error_msg = (
        f"invalid filename, got {filename!r}\n"
        f"filename can't contain any of the following characters: \\/:*?\"<>|"
    )

    invalid_chars = r"[\\/:*?\"<>|]"

    if re.search(invalid_chars, filename):
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg)
        raise ValidationException(error_msg)

    return filename


def valid_image_format(image_format: str) -> str:
    error_msg = (
        f"invalid image format, got {image_format!r}\n"
        f"valid image formats: {C.VALID_IMAGE_FORMATS}"
    )

    image_format = str(image_format).lower().strip(".")
    if image_format not in C.VALID_IMAGE_FORMATS:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg)
        raise ValidationException(error_msg)

    return image_format


def valid_timestamp(timestamp: str, timestamp_type: str) -> str:
    error_msg = f"invalid timestamp, got {timestamp!r}"

    timestamp_as_seconds = None
    if ":" in timestamp:
        regex = r"^([0-9]|[0-5][0-9])(:[0-5][0-9]){1,2}$"
        if not bool(re.match(regex, timestamp)):
            if C.IS_TERMINAL:
                raise argparse.ArgumentTypeError(error_msg)
            raise ValidationException(error_msg)

        timestamp_as_seconds = utils.timestamp_to_seconds(timestamp)

    if timestamp_type == "start":
        timestamp_as_seconds = (
            non_negative_float(timestamp)
            if timestamp_as_seconds is None
            else non_negative_float(timestamp_as_seconds)
        )
    elif timestamp_type == "stop":
        timestamp_as_seconds = (
            positive_float(float(timestamp))
            if timestamp_as_seconds is None
            else positive_float(timestamp_as_seconds)
        )

    return timestamp


def valid_resize_value(resize_value: Union[float, str]) -> float:
    """Video resize mutliplier max value of 50 is arbitrary, but is used to prevent
    the user from accidentally resizing output to abnormally large size.
    """
    error_msg_value = f"invalid resize value, got {resize_value!r}"

    try:
        resize_value = float(resize_value)
    except ValueError:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg_value)
        raise ValidationException(error_msg_value)

    error_msg_range = f"resize value must be between 0.01 and 50, got {resize_value}"
    if not 0.01 <= resize_value <= 50:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg_range)
        raise ValidationException(error_msg_range)

    return resize_value


def valid_dimensions(dimensions: Tuple[int, int]) -> Tuple[int, int]:
    """Dimensions are validated differently when run from command-line because
    argparse's type conversions are applied to each argument individually.
    """
    if len(dimensions) != 2:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(f"invalid dimensions, got {dimensions!r}")
        raise ValidationException(f"invalid dimensions, got {dimensions!r}")

    dimensions = tuple([positive_int(dim) for dim in list(dimensions)])

    return dimensions


def valid_rotate_value(rotate_value: Union[int, str]) -> int:
    """Valid rotate values are 0, 90, 180, 270."""
    error_msg = (
        f"invalid rotate value, got {rotate_value!r}\n"
        f"valid rotate values: {C.VALID_ROTATE_VALUES}"
    )

    try:
        rotate_value = int(rotate_value)
    except ValueError:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg)
        raise ValidationException(error_msg)

    if rotate_value not in C.VALID_ROTATE_VALUES:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg)
        raise ValidationException(error_msg)

    return rotate_value


def valid_start_time(start_time: Union[float, int, str]) -> Union[float, int, str]:
    error_msg = f"invalid start time, got {start_time!r}"

    if isinstance(start_time, int):
        start_time = non_negative_int(start_time)
        return start_time

    if isinstance(start_time, float):
        start_time = non_negative_float(start_time)
        return start_time

    if isinstance(start_time, str):
        start_time = valid_timestamp(start_time, timestamp_type="start")
        return start_time

    if C.IS_TERMINAL:
        raise argparse.ArgumentTypeError(error_msg)
    raise ValidationException(error_msg)


def valid_stop_time(stop_time: Union[float, int, str]) -> Union[float, int, str]:
    error_msg = f"invalid stop time, got {stop_time!r}"

    if isinstance(stop_time, int):
        stop_time = positive_int(stop_time)
        return stop_time

    if isinstance(stop_time, float):
        stop_time = positive_float(stop_time)
        return stop_time

    if isinstance(stop_time, str):
        stop_time = valid_timestamp(stop_time, timestamp_type="stop")
        return stop_time

    if C.IS_TERMINAL:
        raise argparse.ArgumentTypeError(error_msg)
    raise ValidationException(error_msg)


def validate_video_extraction_range(
    start_seconds: float, stop_seconds: float, video_length_seconds: float
) -> None:
    error_msg = f"start time in seconds ({start_seconds}) exceeds video length in seconds ({video_length_seconds})"
    if start_seconds > video_length_seconds:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg)
        raise ValidationException(error_msg)

    error_msg = f"start time in seconds ({start_seconds}) exceeds stop time in seconds ({stop_seconds})"
    if start_seconds > stop_seconds:
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg)
        raise ValidationException(error_msg)


def valid_capture_rate(capture_rate: int, start_frame: float, stop_frame: float) -> int:
    """Capture rate is validated after determining the start and stop video frames."""
    error_msg = (
        f"capture rate ({capture_rate}) exceeds range between "
        f"start ({start_frame}) and stop ({stop_frame}) frames"
    )

    if capture_rate > (stop_frame - start_frame):
        if C.IS_TERMINAL:
            raise argparse.ArgumentTypeError(error_msg)
        raise ValidationException(error_msg)

    return capture_rate



import datetime


def timestamp_to_seconds(timestamp: str) -> int:
    if ":" in timestamp:
        seconds = 0
        for value in timestamp.split(":"):
            seconds = seconds * 60 + int(value)
    else:
        seconds = int(timestamp)

    return seconds


def seconds_to_timestamp(seconds: int) -> str:
    return str(datetime.timedelta(seconds=seconds))
